/*
    Project: TCGINA
    Version: V1.21
    File   : pwdlg.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include <windows.h>
#include <tchar.h>
#include "pwdlg.h"
#include "misc.h"

#define REALTIME_PASSWORD_VERIFICATION
#define SHOW_PASSWORD_QUALITY

#define REFRESH_RATE 100

#if defined(NO_PASSWORD_CONFIRMATION_DIALOG) && defined(SHOW_PASSWORD_QUALITY)
#  undef SHOW_PASSWORD_QUALITY
#endif

#ifdef SHOW_PASSWORD_QUALITY
#  include <tchar.h>
#  include <math.h>
#  ifndef M_LOG2E
#    define M_LOG2E 1.44269504088896340736
#  endif
#endif

static BOOL VerifyPassword(PPASSWORD_DIALOG_DATA ppdd,
  DWORD *pdwPasswordLength, DWORD *pcbPasswordSize)
{
    DWORD dwTempPasswordLength;
    DWORD cbTempPasswordSize;

    if (pdwPasswordLength == NULL) {
        pdwPasswordLength = &dwTempPasswordLength;
    }
    if (pcbPasswordSize == NULL) {
        pcbPasswordSize = &cbTempPasswordSize;
    }

    *pdwPasswordLength = GetWindowTextW(ppdd->hwndPassword,
      ppdd->pwszTempPassword, ppdd->dwMaxPasswordLength + 1);

    if (ppdd->uCodePage == CP_UTF16) {
        *pcbPasswordSize = (*pdwPasswordLength + 1)*sizeof(TCHAR);
    } else {
        *pcbPasswordSize = (DWORD) WideCharToMultiByte(ppdd->uCodePage, 0,
          ppdd->pwszTempPassword, *pdwPasswordLength + 1, NULL, 0, NULL, NULL);
    }

#ifndef NO_PASSWORD_CONFIRMATION_DIALOG
    if (ppdd->hwndConfirmation != NULL) {
         DWORD dwConfirmationLength = GetWindowTextW(ppdd->hwndConfirmation,
          ppdd->pwszConfirmation, ppdd->dwMaxPasswordLength + 1);
        if (dwConfirmationLength != *pdwPasswordLength) {
            return FALSE;
        }
        if (memcmp(ppdd->pwszTempPassword, ppdd->pwszConfirmation,
          dwConfirmationLength * sizeof(TCHAR)) != 0) {
            return FALSE;
        }
    }
#endif

    if (*pdwPasswordLength < ppdd->dwMinPasswordLength) {
        return FALSE;
    }
    if (*pdwPasswordLength == 0) {
        return TRUE;
    }

    return *pcbPasswordSize != 0 && *pcbPasswordSize <= ppdd->cbPasswordSize;
}

INT_PTR CALLBACK PasswordDialogProc(HWND hwndDlg, UINT uMsg,
  WPARAM wParam, LPARAM lParam)
{
    PPASSWORD_DIALOG_DATA ppdd;

    switch (uMsg) {
    case WM_INITDIALOG:
        ppdd = (PPASSWORD_DIALOG_DATA) lParam;
        ppdd->hwndOk               = GetDlgItem(hwndDlg, IDOK);
        ppdd->hwndPassword         = GetDlgItem(hwndDlg, IDC_PASSWORD);
#ifndef NO_PASSWORD_CONFIRMATION_DIALOG
        ppdd->hwndConfirmation     = GetDlgItem(hwndDlg, IDC_CONFIRMATION);
        ppdd->hwndQuality          = GetDlgItem(hwndDlg, IDC_QUALITY);
#endif
        ppdd->hwndNumLock          = GetDlgItem(hwndDlg, IDC_NUMLOCK);
        ppdd->hwndCapsLock         = GetDlgItem(hwndDlg, IDC_CAPSLOCK);
        ppdd->hwndLang             = GetDlgItem(hwndDlg, IDC_LANG);
        ppdd->hwndDisplayPassword  = GetDlgItem(hwndDlg, IDC_DISPLAY_PASSWORD);

        if (ppdd->hwndDisplayPassword) {
            ppdd->wPasswordCharacter = (WPARAM)SendMessage(ppdd->hwndPassword,
              EM_GETPASSWORDCHAR, 0, 0);
            if (ppdd->wPasswordCharacter == 0) {
                ppdd->wPasswordCharacter = (WPARAM)'*';
            }
            if (ppdd->fDisplayPassword) {
                SendMessage(ppdd->hwndDisplayPassword, BM_SETCHECK, BST_CHECKED, 0);
                SendMessage(ppdd->hwndPassword, EM_SETPASSWORDCHAR, 0, 0);
            }
        }

        if (ppdd->pPassword == NULL  || ppdd->cbPasswordSize == 0 ||
#ifndef NO_PASSWORD_CONFIRMATION_DIALOG
          (ppdd->hwndConfirmation != NULL && ppdd->pwszConfirmation == NULL) ||
#endif
          ppdd->pwszTempPassword == NULL || ppdd->dwMaxPasswordLength == 0 ||
          ppdd->hwndOk == NULL || ppdd->hwndPassword == NULL) {
            EndDialog(hwndDlg, IDCANCEL);
            return TRUE;
        }
        ppdd->dwPrevKeyboardStatus = (unsigned)-1;
        ppdd->wPrevLangId          = 0;
        ppdd->dwFirstStartTime     = GetTickCount();
        ppdd->dwStartTime          = ppdd->dwFirstStartTime;
        if (ppdd->hwndLang) {
            HKL hkl;
            if (GetKeyboardLayoutList(0, &hkl) >= 2) {
                ShowWindow(ppdd->hwndLang, SW_SHOWNORMAL);
            } else {
                ShowWindow(ppdd->hwndLang, SW_HIDE);
                ppdd->hwndLang = NULL;
            }
        }
#ifdef REALTIME_PASSWORD_VERIFICATION
        ppdd->fRealTimeVerification = TRUE;
#endif
        SendMessage(ppdd->hwndPassword, EM_LIMITTEXT,
            ppdd->dwMaxPasswordLength, 0);
#ifndef NO_PASSWORD_CONFIRMATION_DIALOG
        if (ppdd->hwndConfirmation) {
            SendMessage(ppdd->hwndConfirmation, EM_LIMITTEXT,
              ppdd->dwMaxPasswordLength, 0);
        }
#endif
        if (ppdd->pszCaption) {
            SetWindowText(hwndDlg, ppdd->pszCaption);
        }

        SetWindowUserData(hwndDlg, ppdd);
        PasswordDialogProc(hwndDlg, WM_TIMER, 0, 0);
        PasswordDialogProc(hwndDlg, WM_COMMAND, (WPARAM)(EN_CHANGE<<16),
          (LPARAM)ppdd->hwndPassword);
        SetTimer(hwndDlg, 0, REFRESH_RATE, 0);

        return TRUE;

    case WM_TIMER: {
        DWORD dwKeyboardStatus;

        ppdd = (PPASSWORD_DIALOG_DATA) GetWindowUserData(hwndDlg);
        dwKeyboardStatus = 0;
        if (GetKeyState(VK_NUMLOCK) & 1) {
            dwKeyboardStatus |= 1;
        }
        if (GetKeyState(VK_CAPITAL) & 1) {
            dwKeyboardStatus |= 2;
        }

        if (dwKeyboardStatus != ppdd->dwPrevKeyboardStatus) {
            ppdd->dwPrevKeyboardStatus = dwKeyboardStatus;

            if (ppdd->hwndNumLock) {
                if (dwKeyboardStatus & 1) {
                    ShowWindow(ppdd->hwndNumLock, SW_SHOWNORMAL);
                } else {
                    ShowWindow(ppdd->hwndNumLock, SW_HIDE);
                }
            }

            if (ppdd->hwndCapsLock) {
                if (dwKeyboardStatus & 2) {
                    ShowWindow(ppdd->hwndCapsLock, SW_SHOWNORMAL);
                } else {
                    ShowWindow(ppdd->hwndCapsLock, SW_HIDE);
                }
            }

        }
        if (ppdd->hwndLang) {
            WORD wLangId = LOWORD(GetKeyboardLayout(0));
            if (wLangId != ppdd->wPrevLangId) {
                TCHAR szLangName[40];
                ppdd->wPrevLangId = wLangId;
                if (GetLocaleInfo(LOWORD(GetKeyboardLayout(0)),
                  LOCALE_SISO639LANGNAME, szLangName,
                  countof(szLangName))) {
                    _tcsupr(szLangName);
                } else {
                    *szLangName = TEXT('\0');
                }
                SetWindowText(ppdd->hwndLang, szLangName);
            }
        }
        if (ppdd->dwTotalTimeout) {
            DWORD dwTime = GetTickCount();
            DWORD dwExpirationTime = ppdd->dwTotalTimeout;
            if (dwTime > ppdd->dwFirstStartTime) {
                dwExpirationTime += ppdd->dwFirstStartTime;
            }
            if (dwTime > dwExpirationTime) {
                PostMessage(hwndDlg, WM_COMMAND, IDCANCEL, 0);
                ppdd->dwTimeout = 0;
                ppdd->dwTotalTimeout = 0;
            }
        }
        if (ppdd->dwTimeout) {
            DWORD dwTime = GetTickCount();
            DWORD dwExpirationTime = ppdd->dwTimeout;
            if (dwTime > ppdd->dwStartTime) {
                dwExpirationTime += ppdd->dwStartTime;
            }
            if (dwTime > dwExpirationTime) {
                PostMessage(hwndDlg, WM_COMMAND, IDCANCEL, 0);
                ppdd->dwTimeout = 0;
                ppdd->dwTotalTimeout = 0;
            }
        }
        return TRUE; }

    case WM_CTLCOLORSTATIC:
        ppdd = (PPASSWORD_DIALOG_DATA) GetWindowUserData(hwndDlg);
        if ((HWND)lParam == ppdd->hwndLang) {
            HDC hDC = (HDC)wParam;
            SetTextColor(hDC, GetSysColor(COLOR_CAPTIONTEXT));
            SetBkMode(hDC, TRANSPARENT);
            return (LRESULT)GetSysColorBrush(COLOR_ACTIVECAPTION);
        }
        break;

    case WM_COMMAND:
        switch (HIWORD(wParam)) {
        case EN_CHANGE:
            ppdd = (PPASSWORD_DIALOG_DATA) GetWindowUserData(hwndDlg);
            ppdd->dwStartTime = GetTickCount();
#ifdef REALTIME_PASSWORD_VERIFICATION
            if (ppdd->fRealTimeVerification) {
                DWORD dwPasswordLength, cbPasswordSize;
                EnableWindow(ppdd->hwndOk, VerifyPassword(ppdd,
                  &dwPasswordLength, &cbPasswordSize));
#  ifdef SHOW_PASSWORD_QUALITY
                if (ppdd->hwndQuality && (HWND)lParam == ppdd->hwndPassword) {
                    TCHAR szQuality[40];
                    BOOL fNonAscii = FALSE;
                    BOOL fSpace = FALSE;
                    BOOL fDigits = FALSE;
                    BOOL fUpperCaseLetters = FALSE;
                    BOOL fLowerCaseLetters = FALSE;
                    BOOL fSymbols = FALSE;
                    DWORD dwQuality;
                    LPWSTR pwszTemp;
                    for (pwszTemp = ppdd->pwszTempPassword; *pwszTemp;
                      pwszTemp++) {
                        WORD w = (WORD)*pwszTemp;
                        if (w < L'\x20' || w > L'\x7e') {
                            fNonAscii = TRUE;
                            break;
                        } else if (w == L' '){
                            fSpace = TRUE;
                        } else if (w >= L'0' && w <= L'9') {
                            fDigits = TRUE;
                        } else if (w >= L'A' && w <= L'Z') {
                            fUpperCaseLetters = TRUE;
                        } else if (w >= L'a' && w <= L'z') {
                            fLowerCaseLetters = TRUE;
                        } else {
                            fSymbols = TRUE;
                        }
                    }
                    if (fNonAscii) {
                        _tcscpy(szQuality, TEXT("NON-ASCII"));
                        dwQuality = 1;
                    } else {
                        DWORD dwCharSetLength = fSpace ? 1 : 0;
                        if (fDigits)           dwCharSetLength += 10;
                        if (fUpperCaseLetters) dwCharSetLength += 26;
                        if (fLowerCaseLetters) dwCharSetLength += 26;
                        if (fSymbols)          dwCharSetLength += 32;
                        dwQuality = (DWORD) (
                          (pwszTemp - ppdd->pwszTempPassword) *
                          (log(dwCharSetLength) * M_LOG2E) + .5);
                        _sntprintf(szQuality, countof(szQuality),
                          TEXT("%u bits"), dwQuality);
                    }
                    SetWindowText(ppdd->hwndQuality, szQuality);
                    ShowWindow(ppdd->hwndQuality, dwQuality ?
                      SW_SHOWNORMAL : SW_HIDE);
                }
#  endif
#  ifndef NO_PASSWORD_CONFIRMATION_DIALOG
                if (ppdd->hwndConfirmation) {
                    BOOL fDisableConfirmationEdit =
                      dwPasswordLength < ppdd->dwMinPasswordLength ||
                      dwPasswordLength > ppdd->dwMaxPasswordLength ||
                      cbPasswordSize == 0 ||
                      cbPasswordSize > ppdd->cbPasswordSize;
                    if (fDisableConfirmationEdit && GetFocus() ==
                      ppdd->hwndConfirmation) {
                        if (dwPasswordLength) {
                            SendMessage(ppdd->hwndPassword, EM_SETSEL, 0,
                            dwPasswordLength);
                        }
                        SetFocus(ppdd->hwndPassword);
                    }
                    EnableWindow(ppdd->hwndConfirmation,
                      !fDisableConfirmationEdit);
                }
#  endif
            }
#endif
            return TRUE;
        }
        switch (wParam) {
        case IDC_DISPLAY_PASSWORD: {
            ppdd = (PPASSWORD_DIALOG_DATA) GetWindowUserData(hwndDlg);
            if (ppdd->hwndDisplayPassword) {
                SendMessage(ppdd->hwndPassword, EM_SETPASSWORDCHAR,
                  (SendMessage(ppdd->hwndDisplayPassword, BM_GETCHECK, 0, 0) ==
                  BST_CHECKED) ? 0 : ppdd->wPasswordCharacter, 0);
                InvalidateRect(ppdd->hwndPassword, NULL, TRUE);
            }
            return TRUE; }

        case IDOK: {
            DWORD dwPasswordLength;
            ppdd = (PPASSWORD_DIALOG_DATA) GetWindowUserData(hwndDlg);
            if (!VerifyPassword(ppdd, &dwPasswordLength, NULL)) {
                if (dwPasswordLength) {
                    SendMessage(ppdd->hwndPassword, EM_SETSEL, 0,
                      dwPasswordLength);
                }
                SetFocus(ppdd->hwndPassword);
#ifdef REALTIME_PASSWORD_VERIFICATION
                EnableWindow(ppdd->hwndOk, FALSE);
#endif
                return TRUE;
            }
            if (ppdd->uCodePage == CP_UTF16) {
                memcpy(ppdd->pPassword, ppdd->pwszTempPassword,
                  (dwPasswordLength+1)*sizeof(TCHAR));
            } else {
                WideCharToMultiByte(ppdd->uCodePage, 0, ppdd->pwszTempPassword,
                  dwPasswordLength + 1, ppdd->pPassword, ppdd->cbPasswordSize,
                  NULL, NULL);
            }

            if (dwPasswordLength) {
                LPWSTR pwszWipePattern;
                LPWSTR pwszWipePatternEnd;
                ppdd = (PPASSWORD_DIALOG_DATA) GetWindowUserData(hwndDlg);
                pwszWipePattern = ppdd->pwszTempPassword;
                pwszWipePatternEnd = pwszWipePattern + dwPasswordLength;
                while (pwszWipePattern != pwszWipePatternEnd) {
                    *pwszWipePattern++ = L' ';
                }
                *pwszWipePattern = L'\0';
#ifdef REALTIME_PASSWORD_VERIFICATION
                ppdd->fRealTimeVerification = FALSE;
#endif
                SetWindowTextW(ppdd->hwndPassword, ppdd->pwszTempPassword);
#ifndef NO_PASSWORD_CONFIRMATION_DIALOG
                if (ppdd->hwndConfirmation) {
                    SetWindowTextW(ppdd->hwndConfirmation,
                      ppdd->pwszTempPassword);
                }
#endif
            }

            /* fall through */ }

        case IDCANCEL: {
            LPWSTR pwszWipePattern;
            LPWSTR pwszWipePatternEnd;
            ppdd = (PPASSWORD_DIALOG_DATA) GetWindowUserData(hwndDlg);
            pwszWipePattern = ppdd->pwszTempPassword;
            pwszWipePatternEnd = pwszWipePattern + ppdd->dwMaxPasswordLength;
            while (pwszWipePattern != pwszWipePatternEnd) {
                *pwszWipePattern++ = L' ';
            }
            *pwszWipePattern = L'\0';
#ifdef REALTIME_PASSWORD_VERIFICATION
            ppdd->fRealTimeVerification = FALSE;
#endif
            SetWindowTextW(ppdd->hwndPassword, ppdd->pwszTempPassword);
#ifndef NO_PASSWORD_CONFIRMATION_DIALOG
            if (ppdd->hwndConfirmation) {
                SetWindowTextW(ppdd->hwndConfirmation, ppdd->pwszTempPassword);
            }
#endif
            KillTimer(hwndDlg, 0);
            if (ppdd->hwndDisplayPassword) {
                if (SendMessage(ppdd->hwndDisplayPassword, BM_GETCHECK, 0,
                  0) == BST_CHECKED) {
                    SendMessage(ppdd->hwndDisplayPassword, BM_SETCHECK,
                      BST_UNCHECKED, 0);
                    SendMessage(ppdd->hwndPassword, EM_SETPASSWORDCHAR,
                      ppdd->wPasswordCharacter, 0);
                }
            }
            EndDialog(hwndDlg, LOWORD(wParam));
            return TRUE; }
        }
        break;
    }

    return FALSE;
}
