/*
    Project: TCGINA
    Version: V1.21
    File   : file.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include <windows.h>
#include <tchar.h>
#include "setup.h"

BOOL FileExists(LPCTSTR pszFileName)
{
    HANDLE hFindFile;
    WIN32_FIND_DATA fd;

    hFindFile = FindFirstFile(pszFileName, &fd);
    if (hFindFile == INVALID_HANDLE_VALUE) {
        HANDLE hFile;
        hFile = CreateFile(pszFileName, 0, FILE_SHARE_READ|FILE_SHARE_WRITE,
          NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
        if (hFile != INVALID_HANDLE_VALUE) {
            CloseHandle(hFile);
            return TRUE;
        } else if (GetLastError() == ERROR_ACCESS_DENIED) {
            return TRUE;
        }

        return FALSE;
    }

    FindClose(hFindFile);

    return TRUE;
}

BOOL UninstallFile(LPCTSTR pszFileName, BOOL *fRebootRequired)
{
    if (!FileExists(pszFileName)) {
        return TRUE;
    }
    DeleteFile(pszFileName);
    if (!FileExists(pszFileName)) {
        return TRUE;
    }
    if (MoveFileEx(pszFileName, NULL, MOVEFILE_DELAY_UNTIL_REBOOT)) {
        if (fRebootRequired) {
            *fRebootRequired = TRUE;
        }
        return TRUE;
    }

    return FileExists(pszFileName);
}

BOOL IsTcginaDLL(PFILE_INFO pfi)
{
    if (*pfi->szInfo) {
        return _tcsncmp(pfi->szInfo, TEXT("TCGINA.DLL;"), 11) == 0;
    } else {
        if (_tcsicmp(pfi->szOriginalFileName, TEXT("TCGINA.DLL")) == 0) {
            if (*pfi->szCompanyName == '\0') {
                return TRUE;
            }
        }
    }

    return FALSE;
}
