/*
    Project: TCGINA
    Version: V1.21
    File   : tcgauto.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tcgpriv.h"
#include "pwdlg.h"
#include "secmem.h"
#include <tchar.h>

/*-----------------------------------------------------------------------------
// Registry function                                                         */

typedef struct
{
    LPCTSTR      pszName;
    PTCG_CONTEXT pCtx;
} TCG_REGENUM_DATA, *PTCG_REGENUM_DATA;

typedef BOOL (CALLBACK * REGENUMPROC)(PTCG_REGENUM_DATA);
static BOOL WINAPI TCG_EnumRegKey(PTCG_CONTEXT pCtx, LPCTSTR pszKeyNameList,
  REGENUMPROC pfRegEnumProc)
{
    TCHAR szKeyNames[TCG_MAX_STRING_LIST];
    LPCTSTR pszCurrName;
    TCG_REGENUM_DATA trd;

    TC_Assert(pszKeyNameList);
    TC_Assert(pfRegEnumProc);

    *szKeyNames = TEXT('\0');
    TC_GetRegistryStringList(HKEY_LOCAL_MACHINE, TCG_REGKEY_NAME,
      pszKeyNameList, szKeyNames, countof(szKeyNames));
    pszCurrName = szKeyNames;

    trd.pCtx = pCtx;
    for (pszCurrName = szKeyNames; *pszCurrName;
      pszCurrName += _tcslen(pszCurrName) + 1) {
        trd.pszName = pszCurrName;
        if (!pfRegEnumProc(&trd)) return FALSE;
    }

    return TRUE;
}

/*-----------------------------------------------------------------------------
// Auto mount                                                                */

#ifdef SUPPORT_AUTOMOUNT

static BOOL CALLBACK TCG_AutomountCallback(PTCG_REGENUM_DATA ptrd)
{
    TCG_SETTINGS  ts;
    TC_HANDLE     hTC;
    PTCG_KEY_DATA ptkd;
    BOOL          fTryDefaultPassword;
    BOOL          fResult;

    TC_Assert(ptrd);
    TC_Assert(ptrd->pCtx);
    TC_Assert(ptrd->pszName);

    fResult = TRUE;

    ptkd = sec_calloc(1, sizeof(*ptkd));
    if (ptkd == NULL) {
        return TRUE;
    }

    TCG_ReadSettings(&ts, ptrd->pszName);
    TC_CopyStringToUTF8(ptkd->aszPassword,
      sizeof(ptkd->aszPassword), ts.szKFPassword);

#ifdef SUPPORT_NETWORK
    if (TC_DriveLetterIsFree(*ts.szNDLocalName)) {
        TCG_MapNetworkDrive(ts.szNDLocalName,
          ts.szNDRemoteName, ts.szNDUserName,
          ts.szNDPassword);
    }
#endif

#ifdef SUPPORT_KEYFILES
    if (!TCG_ReadKeyFile(ptrd->pCtx, ptkd, &ts)) {
        sec_free(ptkd);
        return TRUE;
    }
    if (*ptkd->kdDecryptedData.wszHostName) {
        TC_CopyString(ts.szTCFileName, countof(ts.szTCFileName),
          ptkd->kdDecryptedData.wszHostName);
    }
#endif

#ifdef SUPPORT_TCPASSWORD
    if (*ts.szTCPassword) {
        TC_CopyStringToAnsi(ptkd->aszPassword, sizeof(ptkd->aszPassword),
          ts.szTCPassword);
    }
#endif

    fTryDefaultPassword = *ptkd->aszPassword != '\0';
    if (ptkd->kdDecryptedData.wPasswordSize) {
        fTryDefaultPassword =
          (ptkd->kdDecryptedData.wPasswordType & 3) != KFIPT_INCOMPLETE;
    }

    hTC = TC_Open(0);
    if (hTC) {
        PASSWORD_DIALOG_DATA pdd;
        memset(&pdd, '\0', sizeof(pdd));
        if (ts.dwTCOptions & TC_DISPLAY_PASSWORD) {
            pdd.fDisplayPassword = TRUE;
        }
        pdd.pszCaption = TEXT("Enter Volume Password");
        pdd.uCodePage = CP_ACP;
        pdd.pPassword = ptkd->aszPassword;
        pdd.cbPasswordSize = TC_MAX_PASSWORD;
        pdd.pwszTempPassword = ptkd->tmp.wszTempPassword;
        pdd.dwMinPasswordLength = 1;
        pdd.dwMaxPasswordLength = TC_MAX_PASSWORD_LEN;
        pdd.dwTimeout = ts.dwTCPasswordTimeout;
        pdd.dwTotalTimeout = ts.dwTCPasswordTotalTimeout;
        if (pdd.dwTotalTimeout == 0) {
            pdd.dwTotalTimeout = TCG_AUTOMOUNT_DEFAULT_TIMEOUT;
        }

        for (;;) {

            if (TC_VolumeIsAccessable(*ts.szTCDrive)) {
                TC_DbgPrintf(
                  TEXT("Drive \"%s\" for auto mount volume\"%s\" is already in use!\n"),
                    ts.szTCDrive, ts.szTCFileName);
                break;
            }

            /* Try cached passwords */
            if (TC_MountWithAnsiPassword(hTC, ts.szTCFileName, NULL, ts.szTCDrive,
              ts.dwTCMountOptions, NULL, FALSE, NULL, NULL)) {
#ifdef SUPPORT_SERVICES
                TCG_StopServices(ts.szTCStopServices);
                TCG_StartServices(ts.szTCStartServices);
#endif
                break;
            }

            if (fTryDefaultPassword) {
                if (TCG_Mount(hTC, ts.szTCFileName, ts.szTCDrive, NULL, NULL,
                  ptkd, ts.dwTCMountOptions,
                  NULL, FALSE, NULL)) {
#ifdef SUPPORT_SERVICES
                    TCG_StopServices(ts.szTCStopServices);
                    TCG_StartServices(ts.szTCStartServices);
#endif
                    break;
                }
            }

            for (;;) {
                if (ptrd->pCtx->pWlxFuncs->WlxDialogBoxParam(
                  ptrd->pCtx->hWlx, g_hinstDLL, MAKEINTRESOURCE(IDD_PASSWORD),
                  NULL, PasswordDialogProc, (LPARAM)&pdd) != IDOK) {
                    if ((ts.dwTCOptions &
                      TC_AUTOMOUNT_VOLUME_IS_MANDATORY) == 0) {
                        break;
                    }
                } else if (*ptkd->aszPassword) {
                    if (TCG_Mount(hTC, ts.szTCFileName, ts.szTCDrive,
                      NULL, NULL, ptkd, ts.dwTCMountOptions,
                      NULL, FALSE, NULL)) {
#ifdef SUPPORT_SERVICES
                        TCG_StopServices(ts.szTCStopServices);
                        TCG_StartServices(ts.szTCStartServices);
#endif
                        break;
                    } else {
                        DWORD dwErrorCode = TC_GetLastError(hTC);
                        if (dwErrorCode != ERROR_TC_WRONG_PASSWORD &&
                          dwErrorCode != ERROR_TC_INVALID_PASSWORD &&
                          dwErrorCode != ERROR_TC_INVALID_KEYFILES) {
                            break;
                        }
                    }
                }
            }
            break;
        }
        if (pdd.dwTotalTimeout == 0) {
            /* dialog timed out */
            fResult = FALSE;
        }
        TC_Close(hTC);
    }

    TC_ZeroMemory(&ts, sizeof(ts));

    sec_free(ptkd);

    return fResult;
}

BOOL WINAPI TCG_Automount(PTCG_CONTEXT pCtx)
{
    return TCG_EnumRegKey(pCtx, TEXT("Automount"), TCG_AutomountCallback);
}

#endif
